# scheme/scheme_android.py
"""
Represents an Android 12 color scheme, a mapping of color roles to colors.
"""

from typing import Dict
from PyMCUlib.palettes.core_palette import CorePalette


class SchemeAndroid:
    """
    Represents an Android 12 color scheme, a mapping of color roles to colors.
    """

    def __init__(self, props: Dict[str, int]):
        """
        Private constructor for SchemeAndroid class.

        Args:
            props: A dictionary containing all color properties for the scheme.
        """
        self._props = props

    @property
    def color_accent_primary(self) -> int:
        """Color accent primary."""
        return self._props["color_accent_primary"]

    @property
    def color_accent_primary_variant(self) -> int:
        """Color accent primary variant."""
        return self._props["color_accent_primary_variant"]

    @property
    def color_accent_secondary(self) -> int:
        """Color accent secondary."""
        return self._props["color_accent_secondary"]

    @property
    def color_accent_secondary_variant(self) -> int:
        """Color accent secondary variant."""
        return self._props["color_accent_secondary_variant"]

    @property
    def color_accent_tertiary(self) -> int:
        """Color accent tertiary."""
        return self._props["color_accent_tertiary"]

    @property
    def color_accent_tertiary_variant(self) -> int:
        """Color accent tertiary variant."""
        return self._props["color_accent_tertiary_variant"]

    @property
    def text_color_primary(self) -> int:
        """Text color primary."""
        return self._props["text_color_primary"]

    @property
    def text_color_secondary(self) -> int:
        """Text color secondary."""
        return self._props["text_color_secondary"]

    @property
    def text_color_tertiary(self) -> int:
        """Text color tertiary."""
        return self._props["text_color_tertiary"]

    @property
    def text_color_primary_inverse(self) -> int:
        """Text color primary inverse."""
        return self._props["text_color_primary_inverse"]

    @property
    def text_color_secondary_inverse(self) -> int:
        """Text color secondary inverse."""
        return self._props["text_color_secondary_inverse"]

    @property
    def text_color_tertiary_inverse(self) -> int:
        """Text color tertiary inverse."""
        return self._props["text_color_tertiary_inverse"]

    @property
    def color_background(self) -> int:
        """Color background."""
        return self._props["color_background"]

    @property
    def color_background_floating(self) -> int:
        """Color background floating."""
        return self._props["color_background_floating"]

    @property
    def color_surface(self) -> int:
        """Color surface."""
        return self._props["color_surface"]

    @property
    def color_surface_variant(self) -> int:
        """Color surface variant."""
        return self._props["color_surface_variant"]

    @property
    def color_surface_highlight(self) -> int:
        """Color surface highlight."""
        return self._props["color_surface_highlight"]

    @property
    def surface_header(self) -> int:
        """Surface header."""
        return self._props["surface_header"]

    @property
    def under_surface(self) -> int:
        """Under surface."""
        return self._props["under_surface"]

    @property
    def off_state(self) -> int:
        """Off state."""
        return self._props["off_state"]

    @property
    def accent_surface(self) -> int:
        """Accent surface."""
        return self._props["accent_surface"]

    @property
    def text_primary_on_accent(self) -> int:
        """Text primary on accent."""
        return self._props["text_primary_on_accent"]

    @property
    def text_secondary_on_accent(self) -> int:
        """Text secondary on accent."""
        return self._props["text_secondary_on_accent"]

    @property
    def volume_background(self) -> int:
        """Volume background."""
        return self._props["volume_background"]

    @property
    def scrim(self) -> int:
        """Scrim."""
        return self._props["scrim"]

    @staticmethod
    def light(argb: int) -> 'SchemeAndroid':
        """
        Creates a light Android color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Light Android color scheme, based on the color's hue.
        """
        core = CorePalette.of(argb)
        return SchemeAndroid.light_from_core_palette(core)

    @staticmethod
    def dark(argb: int) -> 'SchemeAndroid':
        """
        Creates a dark Android color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Dark Android color scheme, based on the color's hue.
        """
        core = CorePalette.of(argb)
        return SchemeAndroid.dark_from_core_palette(core)

    @staticmethod
    def light_content(argb: int) -> 'SchemeAndroid':
        """
        Creates a light Android content color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Light Android content color scheme, based on the color's hue.
        """
        core = CorePalette.content_of(argb)
        return SchemeAndroid.light_from_core_palette(core)

    @staticmethod
    def dark_content(argb: int) -> 'SchemeAndroid':
        """
        Creates a dark Android content color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Dark Android content color scheme, based on the color's hue.
        """
        core = CorePalette.content_of(argb)
        return SchemeAndroid.dark_from_core_palette(core)

    @staticmethod
    def light_from_core_palette(core: CorePalette) -> 'SchemeAndroid':
        """
        Creates a light scheme from a core palette.

        Args:
            core: A CorePalette.

        Returns:
            A light Android scheme.
        """
        return SchemeAndroid({
            "color_accent_primary": core.a1.tone(90),
            "color_accent_primary_variant": core.a1.tone(40),
            "color_accent_secondary": core.a2.tone(90),
            "color_accent_secondary_variant": core.a2.tone(40),
            "color_accent_tertiary": core.a3.tone(90),
            "color_accent_tertiary_variant": core.a3.tone(40),
            "text_color_primary": core.n1.tone(10),
            "text_color_secondary": core.n2.tone(30),
            "text_color_tertiary": core.n2.tone(50),
            "text_color_primary_inverse": core.n1.tone(95),
            "text_color_secondary_inverse": core.n1.tone(80),
            "text_color_tertiary_inverse": core.n1.tone(60),
            "color_background": core.n1.tone(95),
            "color_background_floating": core.n1.tone(98),
            "color_surface": core.n1.tone(98),
            "color_surface_variant": core.n1.tone(90),
            "color_surface_highlight": core.n1.tone(100),
            "surface_header": core.n1.tone(90),
            "under_surface": core.n1.tone(0),
            "off_state": core.n1.tone(20),
            "accent_surface": core.a2.tone(95),
            "text_primary_on_accent": core.n1.tone(10),
            "text_secondary_on_accent": core.n2.tone(30),
            "volume_background": core.n1.tone(25),
            "scrim": core.n1.tone(80),
        })

    @staticmethod
    def dark_from_core_palette(core: CorePalette) -> 'SchemeAndroid':
        """
        Creates a dark scheme from a core palette.

        Args:
            core: A CorePalette.

        Returns:
            A dark Android scheme.
        """
        return SchemeAndroid({
            "color_accent_primary": core.a1.tone(90),
            "color_accent_primary_variant": core.a1.tone(70),
            "color_accent_secondary": core.a2.tone(90),
            "color_accent_secondary_variant": core.a2.tone(70),
            "color_accent_tertiary": core.a3.tone(90),
            "color_accent_tertiary_variant": core.a3.tone(70),
            "text_color_primary": core.n1.tone(95),
            "text_color_secondary": core.n2.tone(80),
            "text_color_tertiary": core.n2.tone(60),
            "text_color_primary_inverse": core.n1.tone(10),
            "text_color_secondary_inverse": core.n1.tone(30),
            "text_color_tertiary_inverse": core.n1.tone(50),
            "color_background": core.n1.tone(10),
            "color_background_floating": core.n1.tone(10),
            "color_surface": core.n1.tone(20),
            "color_surface_variant": core.n1.tone(30),
            "color_surface_highlight": core.n1.tone(35),
            "surface_header": core.n1.tone(30),
            "under_surface": core.n1.tone(0),
            "off_state": core.n1.tone(20),
            "accent_surface": core.a2.tone(95),
            "text_primary_on_accent": core.n1.tone(10),
            "text_secondary_on_accent": core.n2.tone(30),
            "volume_background": core.n1.tone(25),
            "scrim": core.n1.tone(80),
        })

    def to_json(self) -> Dict[str, int]:
        """
        Convert the scheme to a JSON serializable dictionary.

        Returns:
            A dictionary representation of the scheme.
        """
        return {**self._props}


# color_spec_2021.py

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor
    from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme

from PyMCUlib.dislike.dislike_analyzer import DislikeAnalyzer
from PyMCUlib.hct.hct import Hct
from PyMCUlib.dynamiccolor.color_spec_delegate import ColorSpecDelegate
from PyMCUlib.dynamiccolor.contrast_curve import ContrastCurve
from PyMCUlib.dynamiccolor.tone_delta_pair import ToneDeltaPair


def _is_fidelity(scheme: 'DynamicScheme') -> bool:
    """
    Returns true if the scheme is Fidelity or Content.
    """
    from .variant import Variant
    return scheme.variant == Variant.FIDELITY or scheme.variant == Variant.CONTENT


def _is_monochrome(scheme: 'DynamicScheme') -> bool:
    """
    Returns true if the scheme is Monochrome.
    """
    from .variant import Variant
    return scheme.variant == Variant.MONOCHROME


def _find_desired_chroma_by_tone(
        hue: float, chroma: float, tone: float,
        by_decreasing_tone: bool) -> float:
    """
    Returns the desired chroma for a given tone at a specific hue.

    Args:
        hue: The given hue.
        chroma: The target chroma.
        tone: The tone to start with.
        by_decreasing_tone: Whether to search for lower tones.
    """
    answer = tone

    closest_to_chroma = Hct.from_hct(hue, chroma, tone)
    if closest_to_chroma.chroma < chroma:
        chroma_peak = closest_to_chroma.chroma
        while closest_to_chroma.chroma < chroma:
            answer += -1.0 if by_decreasing_tone else 1.0
            potential_solution = Hct.from_hct(hue, chroma, answer)
            if chroma_peak > potential_solution.chroma:
                break
            if abs(potential_solution.chroma - chroma) < 0.4:
                break

            potential_delta = abs(potential_solution.chroma - chroma)
            current_delta = abs(closest_to_chroma.chroma - chroma)
            if potential_delta < current_delta:
                closest_to_chroma = potential_solution
            chroma_peak = max(chroma_peak, potential_solution.chroma)

    return answer


class ColorSpecDelegateImpl2021(ColorSpecDelegate):
    """
    A delegate for the dynamic color spec of a DynamicScheme in the 2021 spec.
    """

    ################################################################
    # Main Palettes                                              #
    ################################################################

    def primary_palette_key_color(self) -> 'DynamicColor':
        # Delayed import to avoid circular dependency
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'primary_palette_key_color',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: s.primary_palette.key_color.tone,
        })

    def secondary_palette_key_color(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'secondary_palette_key_color',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: s.secondary_palette.key_color.tone,
        })

    def tertiary_palette_key_color(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'tertiary_palette_key_color',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: s.tertiary_palette.key_color.tone,
        })

    def neutral_palette_key_color(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'neutral_palette_key_color',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: s.neutral_palette.key_color.tone,
        })

    def neutral_variant_palette_key_color(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'neutral_variant_palette_key_color',
            'palette': lambda s: s.neutral_variant_palette,
            'tone': lambda s: s.neutral_variant_palette.key_color.tone,
        })

    def error_palette_key_color(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'error_palette_key_color',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: s.error_palette.key_color.tone,
        })

    ################################################################
    # Surfaces [S]                                               #
    ################################################################

    def background(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'background',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 6 if s.is_dark else 98,
            'is_background': True,
        })

    def on_background(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_background',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 90 if s.is_dark else 10,
            'background': lambda s: self.background(),
            'contrast_curve': lambda s: ContrastCurve(3, 3, 4.5, 7),
        })

    def surface(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 6 if s.is_dark else 98,
            'is_background': True,
        })

    def surface_dim(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_dim',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 6 if s.is_dark else ContrastCurve(87, 87, 80, 75).get(s.contrast_level),
            'is_background': True,
        })

    def surface_bright(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_bright',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(24, 24, 29, 34).get(s.contrast_level) if s.is_dark else 98,
            'is_background': True,
        })

    def surface_container_lowest(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_container_lowest',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(4, 4, 2, 0).get(s.contrast_level) if s.is_dark else 100,
            'is_background': True,
        })

    def surface_container_low(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_container_low',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(10, 10, 11, 12).get(s.contrast_level) if s.is_dark else ContrastCurve(96, 96, 96, 95).get(s.contrast_level),
            'is_background': True,
        })

    def surface_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_container',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(12, 12, 16, 20).get(s.contrast_level) if s.is_dark else ContrastCurve(94, 94, 92, 90).get(s.contrast_level),
            'is_background': True,
        })

    def surface_container_high(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_container_high',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(17, 17, 21, 25).get(s.contrast_level) if s.is_dark else ContrastCurve(92, 92, 88, 85).get(s.contrast_level),
            'is_background': True,
        })

    def surface_container_highest(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_container_highest',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: ContrastCurve(22, 22, 26, 30).get(s.contrast_level) if s.is_dark else ContrastCurve(90, 90, 84, 80).get(s.contrast_level),
            'is_background': True,
        })

    def on_surface(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 90 if s.is_dark else 10,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def surface_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_variant',
            'palette': lambda s: s.neutral_variant_palette,
            'tone': lambda s: 30 if s.is_dark else 90,
            'is_background': True,
        })

    def on_surface_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_surface_variant',
            'palette': lambda s: s.neutral_variant_palette,
            'tone': lambda s: 80 if s.is_dark else 30,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    def inverse_surface(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'inverse_surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 90 if s.is_dark else 20,
            'is_background': True,
        })

    def inverse_on_surface(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'inverse_on_surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 20 if s.is_dark else 95,
            'background': lambda s: self.inverse_surface(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def outline(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'outline',
            'palette': lambda s: s.neutral_variant_palette,
            'tone': lambda s: 60 if s.is_dark else 50,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1.5, 3, 4.5, 7),
        })

    def outline_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'outline_variant',
            'palette': lambda s: s.neutral_variant_palette,
            'tone': lambda s: 30 if s.is_dark else 80,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
        })

    def shadow(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'shadow',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 0,
        })

    def scrim(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'scrim',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 0,
        })

    def surface_tint(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'surface_tint',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 80 if s.is_dark else 40,
            'is_background': True,
        })

    ################################################################
    # Primary [P].                                               #
    ################################################################

    def primary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'primary',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 100 if _is_monochrome(s) and s.is_dark else 0 if _is_monochrome(s) else 80 if s.is_dark else 40,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_container(), self.primary(), 10, 'nearer', False),
        })

    def primary_dim(self) -> 'DynamicColor':
        return None

    def on_primary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_primary',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 10 if _is_monochrome(s) and s.is_dark else 90 if _is_monochrome(s) else 20 if s.is_dark else 100,
            'background': lambda s: self.primary(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def primary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'primary_container',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: s.source_color_hct.tone if _is_fidelity(s) else 85 if _is_monochrome(s) and s.is_dark else 25 if _is_monochrome(s) else 30 if s.is_dark else 90,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_container(), self.primary(), 10, 'nearer', False),
        })

    def on_primary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_primary_container',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: DynamicColor.foreground_tone(
                self.primary_container().tone(s), 4.5
              ) if _is_fidelity(s)
              else 0    if _is_monochrome(s) and s.is_dark
              else 100  if _is_monochrome(s)
              else 90   if s.is_dark
              else 30,
            'background': lambda s: self.primary_container(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    def inverse_primary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'inverse_primary',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 40 if s.is_dark else 80,
            'background': lambda s: self.inverse_surface(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 7),
        })

    #################################################################
    # Secondary [Q].                                              #
    #################################################################

    def secondary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'secondary',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 80 if s.is_dark else 40,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_container(), self.secondary(), 10, 'nearer', False),
        })

    def secondary_dim(self) -> 'DynamicColor':
        return None

    def on_secondary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_secondary',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 10 if _is_monochrome(s) and s.is_dark else 100 if _is_monochrome(s) else 20 if s.is_dark else 100,
            'background': lambda s: self.secondary(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def secondary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'secondary_container',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 30 if _is_monochrome(s) and s.is_dark else 85 if _is_monochrome(s) else 30 if s.is_dark else 90 if not _is_fidelity(s) else _find_desired_chroma_by_tone(
                s.secondary_palette.hue, s.secondary_palette.chroma, 30 if s.is_dark else 90,
                s.is_dark is False),
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_container(), self.secondary(), 10, 'nearer', False),
        })

    def on_secondary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_secondary_container',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 90   if _is_monochrome(s) and s.is_dark
              else 10   if _is_monochrome(s)
              else 90   if s.is_dark
              else 30   if not _is_fidelity(s)
              else DynamicColor.foreground_tone(
                  self.secondary_container().tone(s), 4.5
              ),
            'background': lambda s: self.secondary_container(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Tertiary [T].                                               #
    #################################################################

    def tertiary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'tertiary',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 90 if _is_monochrome(s) and s.is_dark else 25 if _is_monochrome(s) else 80 if s.is_dark else 40,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_container(), self.tertiary(), 10, 'nearer', False),
        })

    def tertiary_dim(self) -> 'DynamicColor':
        return None

    def on_tertiary(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_tertiary',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 10 if _is_monochrome(s) and s.is_dark else 90 if _is_monochrome(s) else 20 if s.is_dark else 100,
            'background': lambda s: self.tertiary(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def tertiary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'tertiary_container',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 60 if _is_monochrome(s) and s.is_dark else 49 if _is_monochrome(s) else 30 if s.is_dark else 90 if not _is_fidelity(s) else DislikeAnalyzer.fix_if_disliked(
                s.tertiary_palette.get_hct(s.source_color_hct.tone)).tone,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_container(), self.tertiary(), 10, 'nearer', False),
        })

    def on_tertiary_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_tertiary_container',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s:  0   if _is_monochrome(s) and s.is_dark
              else 100  if _is_monochrome(s)
              else  90  if s.is_dark
              else  30  if not _is_fidelity(s)
              else DynamicColor.foreground_tone(
                  self.tertiary_container().tone(s), 4.5
              ),
            'background': lambda s: self.tertiary_container(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Error [E].                                                   #
    #################################################################

    def error(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'error',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: 80 if s.is_dark else 40,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.error_container(), self.error(), 10, 'nearer', False),
        })

    def error_dim(self) -> 'DynamicColor':
        return None

    def on_error(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_error',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: 20 if s.is_dark else 100,
            'background': lambda s: self.error(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def error_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'error_container',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: 30 if s.is_dark else 90,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.error_container(), self.error(), 10, 'nearer', False),
        })

    def on_error_container(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_error_container',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: 90   if _is_monochrome(s) and s.is_dark
              else 10   if _is_monochrome(s)
              else 90   if s.is_dark
              else 30,
            'background': lambda s: self.error_container(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Primary Fixed [PF]                                           #
    #################################################################

    def primary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'primary_fixed',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 40.0 if _is_monochrome(s) else 90.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_fixed(), self.primary_fixed_dim(), 10, 'lighter', True),
        })

    def primary_fixed_dim(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'primary_fixed_dim',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 30.0 if _is_monochrome(s) else 80.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_fixed(), self.primary_fixed_dim(), 10, 'lighter', True),
        })

    def on_primary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_primary_fixed',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 100.0 if _is_monochrome(s) else 10.0,
            'background': lambda s: self.primary_fixed_dim(),
            'second_background': lambda s: self.primary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def on_primary_fixed_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_primary_fixed_variant',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: 90.0 if _is_monochrome(s) else 30.0,
            'background': lambda s: self.primary_fixed_dim(),
            'second_background': lambda s: self.primary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Secondary Fixed [QF]                                          #
    #################################################################

    def secondary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'secondary_fixed',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 80.0 if _is_monochrome(s) else 90.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_fixed(), self.secondary_fixed_dim(), 10, 'lighter', True),
        })

    def secondary_fixed_dim(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'secondary_fixed_dim',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 70.0 if _is_monochrome(s) else 80.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_fixed(), self.secondary_fixed_dim(), 10, 'lighter', True),
        })

    def on_secondary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_secondary_fixed',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 10.0,
            'background': lambda s: self.secondary_fixed_dim(),
            'second_background': lambda s: self.secondary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def on_secondary_fixed_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_secondary_fixed_variant',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: 25.0 if _is_monochrome(s) else 30.0,
            'background': lambda s: self.secondary_fixed_dim(),
            'second_background': lambda s: self.secondary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Tertiary Fixed [TF]                                         #
    #################################################################

    def tertiary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'tertiary_fixed',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 40.0 if _is_monochrome(s) else 90.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_fixed(), self.tertiary_fixed_dim(), 10, 'lighter', True),
        })

    def tertiary_fixed_dim(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'tertiary_fixed_dim',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 30.0 if _is_monochrome(s) else 80.0,
            'is_background': True,
            'background': lambda s: self.highest_surface(s),
            'contrast_curve': lambda s: ContrastCurve(1, 1, 3, 4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_fixed(), self.tertiary_fixed_dim(), 10, 'lighter', True),
        })

    def on_tertiary_fixed(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_tertiary_fixed',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 100.0 if _is_monochrome(s) else 10.0,
            'background': lambda s: self.tertiary_fixed_dim(),
            'second_background': lambda s: self.tertiary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(4.5, 7, 11, 21),
        })

    def on_tertiary_fixed_variant(self) -> 'DynamicColor':
        from .dynamic_color import DynamicColor
        return DynamicColor.from_palette({
            'name': 'on_tertiary_fixed_variant',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: 90.0 if _is_monochrome(s) else 30.0,
            'background': lambda s: self.tertiary_fixed_dim(),
            'second_background': lambda s: self.tertiary_fixed(),
            'contrast_curve': lambda s: ContrastCurve(3, 4.5, 7, 11),
        })

    #################################################################
    # Other                                                      #
    #################################################################

    def highest_surface(self, s: 'DynamicScheme') -> 'DynamicColor':
        return self.surface_bright() if s.is_dark else self.surface_dim()


# dynamiccolor/tone_delta_pair.py

from typing import Literal, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .dynamic_color import DynamicColor

# Describes the different in tone between colors.
# nearer and farther are deprecated. Use DeltaConstraint instead.
TonePolarity = Literal['darker', 'lighter', 'nearer', 'farther', 'relative_darker', 'relative_lighter']

# Describes how to fulfill a tone delta pair constraint.
DeltaConstraint = Literal['exact', 'nearer', 'farther']


class ToneDeltaPair:
    """
    Documents a constraint between two DynamicColors, in which their tones must
    have a certain distance from each other.

    Prefer a DynamicColor with a background, this is for special cases when
    designers want tonal distance, literally contrast, between two colors that
    don't have a background / foreground relationship or a contrast guarantee.
    """

    def __init__(
            self,
            role_a: 'DynamicColor',
            role_b: 'DynamicColor',
            delta: float,
            polarity: TonePolarity,
            stay_together: bool,
            constraint: Optional[DeltaConstraint] = None,
    ):
        """
        Documents a constraint in tone distance between two DynamicColors.

        The polarity is an adjective that describes "A", compared to "B".

        For instance, ToneDeltaPair(A, B, 15, 'darker', 'exact') states that
        A's tone should be exactly 15 darker than B's.

        'relative_darker' and 'relative_lighter' describes the tone adjustment
        relative to the surface color trend (white in light mode; black in dark
        mode). For instance, ToneDeltaPair(A, B, 10, 'relative_lighter',
        'farther') states that A should be at least 10 lighter than B in light
        mode, and at least 10 darker than B in dark mode.

        Args:
            role_a: The first role in a pair.
            role_b: The second role in a pair.
            delta: Required difference between tones. Absolute value, negative
                values have undefined behavior.
            polarity: The relative relation between tones of roleA and roleB,
                as described above.
            stay_together: Whether these two roles should stay on the same side
                of the "awkward zone" (T50-59). This is necessary for certain cases where
                one role has two backgrounds.
            constraint: How to fulfill the tone delta pair constraint.
        """
        self.role_a = role_a
        self.role_b = role_b
        self.delta = delta
        self.polarity = polarity
        self.stay_together = stay_together
        self.constraint = constraint if constraint is not None else 'exact'


# dynamiccolor/color_spec_2025.py

from typing import Optional, TYPE_CHECKING
import copy

if TYPE_CHECKING:
    from PyMCUlib.palettes.tonal_palette import TonalPalette
    from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor

from PyMCUlib.hct.hct import Hct
from PyMCUlib.utils import math_utils

from PyMCUlib.dynamiccolor.color_spec_2021 import ColorSpecDelegateImpl2021
from PyMCUlib.dynamiccolor.contrast_curve import ContrastCurve
from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor, extend_spec_version
from PyMCUlib.dynamiccolor.tone_delta_pair import ToneDeltaPair

def _clone_light(scheme):
    # 浅拷贝 scheme 对象，并将 is_dark 强制设为 False（轻色模式）
    new = copy.copy(scheme)
    new.is_dark = False
    return new

def _t_max_c(
        palette: 'TonalPalette', 
        lower_bound: float = 0, 
        upper_bound: float = 100,
        chroma_multiplier: float = 1) -> float:
    """
    Returns the maximum tone for a given chroma in the palette.

    Args:
        palette: The tonal palette to use.
        lower_bound: The lower bound of the tone.
        upper_bound: The upper bound of the tone.
        chroma_multiplier: The multiplier for the chroma.
    """
    answer = _find_best_tone_for_chroma(
        palette.hue, palette.chroma * chroma_multiplier, 100, True)
    return math_utils.clamp_double(lower_bound, upper_bound, answer)


def _t_min_c(
        palette: 'TonalPalette', 
        lower_bound: float = 0,
        upper_bound: float = 100) -> float:
    """
    Returns the minimum tone for a given chroma in the palette.

    Args:
        palette: The tonal palette to use.
        lower_bound: The lower bound of the tone.
        upper_bound: The upper bound of the tone.
    """
    answer = _find_best_tone_for_chroma(palette.hue, palette.chroma, 0, False)
    return math_utils.clamp_double(lower_bound, upper_bound, answer)


def _find_best_tone_for_chroma(
        hue: float, chroma: float, tone: float,
        by_decreasing_tone: bool) -> float:
    """
    Searches for the best tone with a given chroma from a given tone at a
    specific hue.

    Args:
        hue: The given hue.
        chroma: The target chroma.
        tone: The tone to start with.
        by_decreasing_tone: Whether to search for lower tones.
    """
    answer = tone
    best_candidate = Hct.from_hct(hue, chroma, answer)
    
    while best_candidate.chroma < chroma:
        if tone < 0 or tone > 100:
            break
        tone += -1.0 if by_decreasing_tone else 1.0
        new_candidate = Hct.from_hct(hue, chroma, tone)
        if best_candidate.chroma < new_candidate.chroma:
            best_candidate = new_candidate
            answer = tone

    return answer


def _get_curve(default_contrast: float) -> ContrastCurve:
    """
    Returns the contrast curve for a given default contrast.

    Args:
        default_contrast: The default contrast to use.
    """
    if default_contrast == 1.5:
        return ContrastCurve(1.5, 1.5, 3, 4.5)
    elif default_contrast == 3:
        return ContrastCurve(3, 3, 4.5, 7)
    elif default_contrast == 4.5:
        return ContrastCurve(4.5, 4.5, 7, 11)
    elif default_contrast == 6:
        return ContrastCurve(6, 6, 7, 11)
    elif default_contrast == 7:
        return ContrastCurve(7, 7, 11, 21)
    elif default_contrast == 9:
        return ContrastCurve(9, 9, 11, 21)
    elif default_contrast == 11:
        return ContrastCurve(11, 11, 21, 21)
    elif default_contrast == 21:
        return ContrastCurve(21, 21, 21, 21)
    else:
        # Shouldn't happen.
        return ContrastCurve(default_contrast, default_contrast, 7, 21)


class ColorSpecDelegateImpl2025(ColorSpecDelegateImpl2021):
    """
    A delegate for the dynamic color spec of a DynamicScheme in the 2025 spec.
    """

    ################################################################
    # Surfaces [S]                                               #
    ################################################################

    def surface(self) -> 'DynamicColor':
        base = super().surface()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                base.tone(s),
                4 if s.platform == 'phone' and s.is_dark else (
                    99 if s.platform == 'phone' and Hct.is_yellow(s.neutral_palette.hue) else
                    97 if s.platform == 'phone' and s.variant == 'VIBRANT' else
                    98 if s.platform == 'phone' else 0
                )
            )[1],
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_dim(self) -> 'DynamicColor':
        base = super().surface_dim()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_dim',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                4 if s.is_dark else (
                    90 if Hct.is_yellow(s.neutral_palette.hue) else
                    85 if s.variant == 'VIBRANT' else
                    87
                )
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                1 if s.is_dark else (
                    2.5 if s.variant == 'NEUTRAL' else
                    1.7 if s.variant == 'TONAL_SPOT' else
                    2.7 if s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                    1.75 if s.variant == 'EXPRESSIVE' else
                    1.36 if s.variant == 'VIBRANT' else 1
                )
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_bright(self) -> 'DynamicColor':
        base = super().surface_bright()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_bright',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                18 if s.is_dark else (
                    99 if Hct.is_yellow(s.neutral_palette.hue) else
                    97 if s.variant == 'VIBRANT' else
                    98
                )
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                2.5 if s.is_dark and s.variant == 'NEUTRAL' else
                1.7 if s.is_dark and s.variant == 'TONAL_SPOT' else
                2.7 if s.is_dark and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.75 if s.is_dark and s.variant == 'EXPRESSIVE' else
                1.36 if s.is_dark and s.variant == 'VIBRANT' else 1
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_container_lowest(self) -> 'DynamicColor':
        base = super().surface_container_lowest()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_container_lowest',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 0 if s.is_dark else 100,
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_container_low(self) -> 'DynamicColor':
        base = super().surface_container_low()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_container_low',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                6 if s.platform == 'phone' and s.is_dark else
                98 if s.platform == 'phone' and Hct.is_yellow(s.neutral_palette.hue) else
                95 if s.platform == 'phone' and s.variant == 'VIBRANT' else
                96 if s.platform == 'phone' else 15
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                1.3 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.25 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                1.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.15 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else
                1.08 if s.platform == 'phone' and s.variant == 'VIBRANT' else 1
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_container(self) -> 'DynamicColor':
        base = super().surface_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_container',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                9 if s.platform == 'phone' and s.is_dark else
                96 if s.platform == 'phone' and Hct.is_yellow(s.neutral_palette.hue) else
                92 if s.platform == 'phone' and s.variant == 'VIBRANT' else
                94 if s.platform == 'phone' else 20
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                1.6 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.4 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                1.6 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else
                1.15 if s.platform == 'phone' and s.variant == 'VIBRANT' else 1
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_container_high(self) -> 'DynamicColor':
        base = super().surface_container_high()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_container_high',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                12 if s.platform == 'phone' and s.is_dark else
                94 if s.platform == 'phone' and Hct.is_yellow(s.neutral_palette.hue) else
                90 if s.platform == 'phone' and s.variant == 'VIBRANT' else
                92 if s.platform == 'phone' else 25
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                1.9 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.5 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                1.95 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.45 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else
                1.22 if s.platform == 'phone' and s.variant == 'VIBRANT' else 1
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_container_highest(self) -> 'DynamicColor':
        base = super().surface_container_highest()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_container_highest',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                15 if s.is_dark else
                92 if Hct.is_yellow(s.neutral_palette.hue) else
                88 if s.variant == 'VIBRANT' else 90
            ),
            'is_background': True,
            'chroma_multiplier': lambda s: (
                2.2 if s.variant == 'NEUTRAL' else
                1.7 if s.variant == 'TONAL_SPOT' else
                2.3 if s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.6 if s.variant == 'EXPRESSIVE' else
                1.29 if s.variant == 'VIBRANT' else 1
            ),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_surface(self) -> 'DynamicColor':
        base = super().on_surface()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: (
                _t_max_c(s.neutral_palette, 0, 100, 1.1) if s.variant == 'VIBRANT' else
                DynamicColor.get_initial_tone_from_background(
                    lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high()
                )(s)
            ),
            'chroma_multiplier': lambda s: (
                2.2 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.7 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                3.0 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) and s.is_dark else
                2.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.6 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else 1
            ),
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(11) if s.is_dark else _get_curve(9),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_surface_variant(self) -> 'DynamicColor':
        base = super().on_surface_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_surface_variant',
            'palette': lambda s: s.neutral_palette,
            'chroma_multiplier': lambda s: (
                2.2 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.7 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                3.0 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) and s.is_dark else
                2.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.6 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else 1
            ),
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def outline(self) -> 'DynamicColor':
        base = super().outline()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'outline',
            'palette': lambda s: s.neutral_palette,
            'chroma_multiplier': lambda s: (
                2.2 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.7 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                3.0 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) and s.is_dark else
                2.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.6 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else 1
            ),
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(3) if s.platform == 'phone' else _get_curve(4.5),
        })
        return extend_spec_version(base, '2025', color2025)

    def outline_variant(self) -> 'DynamicColor':
        base = super().outline_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'outline_variant',
            'palette': lambda s: s.neutral_palette,
            'chroma_multiplier': lambda s: (
                2.2 if s.platform == 'phone' and s.variant == 'NEUTRAL' else
                1.7 if s.platform == 'phone' and s.variant == 'TONAL_SPOT' else
                3.0 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) and s.is_dark else
                2.3 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' and Hct.is_yellow(s.neutral_palette.hue) else
                1.6 if s.platform == 'phone' and s.variant == 'EXPRESSIVE' else 1
            ),
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(1.5) if s.platform == 'phone' else _get_curve(3),
        })
        return extend_spec_version(base, '2025', color2025)

    def inverse_surface(self) -> 'DynamicColor':
        base = super().inverse_surface()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'inverse_surface',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: 98 if s.is_dark else 4,
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def inverse_on_surface(self) -> 'DynamicColor':
        base = super().inverse_on_surface()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'inverse_on_surface',
            'palette': lambda s: s.neutral_palette,
            'background': lambda s: self.inverse_surface(),
            'contrast_curve': lambda s: _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    ################################################################
    # Primaries [P]                                              #
    ################################################################

    def primary(self) -> 'DynamicColor':
        base = super().primary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'primary',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: (
                90 if s.variant == 'NEUTRAL' and s.platform != 'phone' else
                80 if s.variant == 'NEUTRAL' and s.platform == 'phone' and s.is_dark else
                40 if s.variant == 'NEUTRAL' and s.platform == 'phone' else
                _t_max_c(s.primary_palette, 0, 90) if s.variant == 'TONAL_SPOT' and s.platform != 'phone' else
                80 if s.variant == 'TONAL_SPOT' and s.platform == 'phone' and s.is_dark else
                _t_max_c(s.primary_palette) if s.variant == 'TONAL_SPOT' and s.platform == 'phone' else
                _t_max_c(
                    s.primary_palette, 0,
                    25 if Hct.is_yellow(s.primary_palette.hue) else
                    88 if Hct.is_cyan(s.primary_palette.hue) else 98
                ) if s.variant == 'EXPRESSIVE' else
                _t_max_c(
                    s.primary_palette, 0, 
                    88 if Hct.is_cyan(s.primary_palette.hue) else 98
                ) if s.variant == 'VIBRANT' else
                base.tone(s)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_container(), self.primary(), 5, 'relative_lighter',
                True, 'farther'
            ) if s.platform == 'phone' else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def primary_dim(self) -> Optional['DynamicColor']:
        return DynamicColor.from_palette({
            'name': 'primary_dim',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: (
                85 if s.variant == 'NEUTRAL' else
                _t_max_c(s.primary_palette, 0, 90) if s.variant == 'TONAL_SPOT' else
                _t_max_c(s.primary_palette)
            ),
            'is_background': True,
            'background': lambda s: self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_dim(), self.primary(), 5, 'darker', True, 'farther'),
        })

    def on_primary(self) -> 'DynamicColor':
        base = super().on_primary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_primary',
            'palette': lambda s: s.primary_palette,
            'background': lambda s: self.primary_dim() if s.platform != 'phone' else self.primary(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def primary_container(self) -> 'DynamicColor':
        base = super().primary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'primary_container',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: (
                30 if s.platform == 'watch' else
                30 if s.variant == 'NEUTRAL' and s.is_dark else
                90 if s.variant == 'NEUTRAL' else
                _t_min_c(s.primary_palette, 35, 93) if s.variant == 'TONAL_SPOT' and s.is_dark else
                _t_max_c(s.primary_palette, 0, 90) if s.variant == 'TONAL_SPOT' else
                _t_max_c(s.primary_palette, 30, 93) if s.variant == 'EXPRESSIVE' and s.is_dark else
                _t_max_c(
                    s.primary_palette, 78,
                    88 if Hct.is_cyan(s.primary_palette.hue) else 90
                ) if s.variant == 'EXPRESSIVE' else
                _t_min_c(s.primary_palette, 66, 93) if s.variant == 'VIBRANT' and s.is_dark else
                _t_max_c(
                    s.primary_palette, 66,
                    88 if Hct.is_cyan(s.primary_palette.hue) else 93
                ) if s.variant == 'VIBRANT' else
                base.tone(s)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else None,
            'tone_delta_pair': lambda s: (
                None if s.platform == 'phone' else
                ToneDeltaPair(
                    self.primary_container(), self.primary_dim(), 10, 'darker', True,
                    'farther'
                )
            ),
            'contrast_curve': lambda s: _get_curve(1.5) if s.platform == 'phone' and s.contrast_level > 0 else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def on_primary_container(self) -> 'DynamicColor':
        base = super().on_primary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_primary_container',
            'palette': lambda s: s.primary_palette,
            'background': lambda s: self.primary_container(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def primary_fixed(self) -> 'DynamicColor':
        base = super().primary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'primary_fixed',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: self.primary_container().get_tone(_clone_light(s)),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def primary_fixed_dim(self) -> 'DynamicColor':
        base = super().primary_fixed_dim()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'primary_fixed_dim',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: self.primary_fixed().get_tone(s),
            'is_background': True,
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.primary_fixed_dim(), self.primary_fixed(), 5, 'darker', True,
                'exact'),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_primary_fixed(self) -> 'DynamicColor':
        base = super().on_primary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_primary_fixed',
            'palette': lambda s: s.primary_palette,
            'background': lambda s: self.primary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_primary_fixed_variant(self) -> 'DynamicColor':
        base = super().on_primary_fixed_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_primary_fixed_variant',
            'palette': lambda s: s.primary_palette,
            'background': lambda s: self.primary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(4.5),
        })
        return extend_spec_version(base, '2025', color2025)

    def inverse_primary(self) -> 'DynamicColor':
        base = super().inverse_primary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'inverse_primary',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: _t_max_c(s.primary_palette),
            'background': lambda s: self.inverse_surface(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    ################################################################
    # Secondaries [Q]                                            #
    ################################################################

    def secondary(self) -> 'DynamicColor':
        base = super().secondary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'secondary',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: (
                90 if s.platform == 'watch' and s.variant == 'NEUTRAL' else
                _t_max_c(s.secondary_palette, 0, 90) if s.platform == 'watch' else
                _t_min_c(s.secondary_palette, 0, 98) if s.variant == 'NEUTRAL' and s.is_dark else
                _t_max_c(s.secondary_palette) if s.variant == 'NEUTRAL' else
                _t_max_c(s.secondary_palette, 0, 98 if s.is_dark else 90) if s.variant == 'VIBRANT' else
                80 if s.is_dark else _t_max_c(s.secondary_palette)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_container(), self.secondary(), 5,
                'relative_lighter', True, 'farther'
            ) if s.platform == 'phone' else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def secondary_dim(self) -> Optional['DynamicColor']:
        return DynamicColor.from_palette({
            'name': 'secondary_dim',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: (
                85 if s.variant == 'NEUTRAL' else
                _t_max_c(s.secondary_palette, 0, 90)
            ),
            'is_background': True,
            'background': lambda s: self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_dim(), self.secondary(), 5, 'darker', True, 'farther'),
        })
        
    def on_secondary(self) -> 'DynamicColor':
        base = super().on_secondary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_secondary',
            'palette': lambda s: s.secondary_palette,
            'background': lambda s: self.secondary() if s.platform == 'phone' else self.secondary_dim(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def secondary_container(self) -> 'DynamicColor':
        base = super().secondary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'secondary_container',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: (
                30 if s.platform == 'watch' else
                _t_min_c(s.secondary_palette, 30, 40) if s.variant == 'VIBRANT' and s.is_dark else
                _t_max_c(s.secondary_palette, 84, 90) if s.variant == 'VIBRANT' else
                15 if s.variant == 'EXPRESSIVE' and s.is_dark else
                _t_max_c(s.secondary_palette, 90, 95) if s.variant == 'EXPRESSIVE' else
                25 if s.is_dark else 90
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else None,
            'tone_delta_pair': lambda s: (
                ToneDeltaPair(
                    self.secondary_container(), self.secondary_dim(), 10, 'darker',
                    True, 'farther'
                ) if s.platform == 'watch' else None
            ),
            'contrast_curve': lambda s: _get_curve(1.5) if s.platform == 'phone' and s.contrast_level > 0 else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def on_secondary_container(self) -> 'DynamicColor':
        base = super().on_secondary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_secondary_container',
            'palette': lambda s: s.secondary_palette,
            'background': lambda s: self.secondary_container(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def secondary_fixed(self) -> 'DynamicColor':
        base = super().secondary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'secondary_fixed',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: self.secondary_container().get_tone(_clone_light(s)),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def secondary_fixed_dim(self) -> 'DynamicColor':
        base = super().secondary_fixed_dim()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'secondary_fixed_dim',
            'palette': lambda s: s.secondary_palette,
            'tone': lambda s: self.secondary_fixed().get_tone(s),
            'is_background': True,
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.secondary_fixed_dim(), self.secondary_fixed(), 5, 'darker', True,
                'exact'),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_secondary_fixed(self) -> 'DynamicColor':
        base = super().on_secondary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_secondary_fixed',
            'palette': lambda s: s.secondary_palette,
            'background': lambda s: self.secondary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_secondary_fixed_variant(self) -> 'DynamicColor':
        base = super().on_secondary_fixed_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_secondary_fixed_variant',
            'palette': lambda s: s.secondary_palette,
            'background': lambda s: self.secondary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(4.5),
        })
        return extend_spec_version(base, '2025', color2025)

    ################################################################
    # Tertiaries [T]                                             #
    ################################################################

    def tertiary(self) -> 'DynamicColor':
        base = super().tertiary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'tertiary',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: (
                _t_max_c(s.tertiary_palette, 0, 90) if s.platform == 'watch' and s.variant == 'TONAL_SPOT' else
                _t_max_c(s.tertiary_palette) if s.platform == 'watch' else
                _t_max_c(
                    s.tertiary_palette, 0,
                    88 if Hct.is_cyan(s.tertiary_palette.hue) else (s.is_dark and 98 or 100)
                ) if s.variant in ('EXPRESSIVE', 'VIBRANT') else
                _t_max_c(s.tertiary_palette, 0, 98) if s.variant in ('NEUTRAL', 'TONAL_SPOT') and s.is_dark else
                _t_max_c(s.tertiary_palette)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_container(), self.tertiary(), 5, 'relative_lighter',
                True, 'farther'
            ) if s.platform == 'phone' else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def tertiary_dim(self) -> Optional['DynamicColor']:
        return DynamicColor.from_palette({
            'name': 'tertiary_dim',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: (
                _t_max_c(s.tertiary_palette, 0, 90) if s.variant == 'TONAL_SPOT' else
                _t_max_c(s.tertiary_palette)
            ),
            'is_background': True,
            'background': lambda s: self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_dim(), self.tertiary(), 5, 'darker', True, 'farther'),
        })

    def on_tertiary(self) -> 'DynamicColor':
        base = super().on_tertiary()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_tertiary',
            'palette': lambda s: s.tertiary_palette,
            'background': lambda s: self.tertiary() if s.platform == 'phone' else self.tertiary_dim(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def tertiary_container(self) -> 'DynamicColor':
        base = super().tertiary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'tertiary_container',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: (
                _t_max_c(s.tertiary_palette, 0, 90) if s.platform == 'watch' and s.variant == 'TONAL_SPOT' else
                _t_max_c(s.tertiary_palette) if s.platform == 'watch' else
                _t_max_c(s.tertiary_palette, 0, 93) if s.variant == 'NEUTRAL' and s.is_dark else
                _t_max_c(s.tertiary_palette, 0, 96) if s.variant == 'NEUTRAL' else
                _t_max_c(s.tertiary_palette, 0, 93 if s.is_dark else 100) if s.variant == 'TONAL_SPOT' else
                _t_max_c(
                    s.tertiary_palette, 75,
                    88 if Hct.is_cyan(s.tertiary_palette.hue) else (s.is_dark and 93 or 100)
                ) if s.variant == 'EXPRESSIVE' else
                _t_max_c(s.tertiary_palette, 0, 93) if s.variant == 'VIBRANT' and s.is_dark else
                _t_max_c(s.tertiary_palette, 72, 100)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else None,
            'tone_delta_pair': lambda s: (
                ToneDeltaPair(
                    self.tertiary_container(), self.tertiary_dim(), 10, 'darker', True,
                    'farther'
                ) if s.platform == 'watch' else None
            ),
            'contrast_curve': lambda s: _get_curve(1.5) if s.platform == 'phone' and s.contrast_level > 0 else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def on_tertiary_container(self) -> 'DynamicColor':
        base = super().on_tertiary_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_tertiary_container',
            'palette': lambda s: s.tertiary_palette,
            'background': lambda s: self.tertiary_container(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def tertiary_fixed(self) -> 'DynamicColor':
        base = super().tertiary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'tertiary_fixed',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: self.tertiary_container().get_tone(_clone_light(s)),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def tertiary_fixed_dim(self) -> 'DynamicColor':
        base = super().tertiary_fixed_dim()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'tertiary_fixed_dim',
            'palette': lambda s: s.tertiary_palette,
            'tone': lambda s: self.tertiary_fixed().get_tone(s),
            'is_background': True,
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.tertiary_fixed_dim(), self.tertiary_fixed(), 5, 'darker', True,
                'exact'),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_tertiary_fixed(self) -> 'DynamicColor':
        base = super().on_tertiary_fixed()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_tertiary_fixed',
            'palette': lambda s: s.tertiary_palette,
            'background': lambda s: self.tertiary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def on_tertiary_fixed_variant(self) -> 'DynamicColor':
        base = super().on_tertiary_fixed_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_tertiary_fixed_variant',
            'palette': lambda s: s.tertiary_palette,
            'background': lambda s: self.tertiary_fixed_dim(),
            'contrast_curve': lambda s: _get_curve(4.5),
        })
        return extend_spec_version(base, '2025', color2025)

    ################################################################
    # Errors [E]                                                 #
    ################################################################

    def error(self) -> 'DynamicColor':
        base = super().error()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'error',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: (
                _t_min_c(s.error_palette, 0, 98) if s.platform == 'phone' and s.is_dark else
                _t_max_c(s.error_palette) if s.platform == 'phone' else
                _t_min_c(s.error_palette)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.error_container(), self.error(), 5, 'relative_lighter', True,
                'farther'
            ) if s.platform == 'phone' else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def error_dim(self) -> Optional['DynamicColor']:
        return DynamicColor.from_palette({
            'name': 'error_dim',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: _t_min_c(s.error_palette),
            'is_background': True,
            'background': lambda s: self.surface_container_high(),
            'contrast_curve': lambda s: _get_curve(4.5),
            'tone_delta_pair': lambda s: ToneDeltaPair(
                self.error_dim(), self.error(), 5, 'darker', True, 'farther'),
        })

    def on_error(self) -> 'DynamicColor':
        base = super().on_error()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_error',
            'palette': lambda s: s.error_palette,
            'background': lambda s: self.error() if s.platform == 'phone' else self.error_dim(),
            'contrast_curve': lambda s: _get_curve(6) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    def error_container(self) -> 'DynamicColor':
        base = super().error_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'error_container',
            'palette': lambda s: s.error_palette,
            'tone': lambda s: (
                30 if s.platform == 'watch' else
                _t_min_c(s.error_palette, 30, 93) if s.is_dark else
                _t_max_c(s.error_palette, 0, 90)
            ),
            'is_background': True,
            'background': lambda s: self.highest_surface(s) if s.platform == 'phone' else None,
            'tone_delta_pair': lambda s: (
                ToneDeltaPair(
                    self.error_container(), self.error_dim(), 10, 'darker', True,
                    'farther'
                ) if s.platform == 'watch' else None
            ),
            'contrast_curve': lambda s: _get_curve(1.5) if s.platform == 'phone' and s.contrast_level > 0 else None,
        })
        return extend_spec_version(base, '2025', color2025)

    def on_error_container(self) -> 'DynamicColor':
        base = super().on_error_container()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_error_container',
            'palette': lambda s: s.error_palette,
            'background': lambda s: self.error_container(),
            'contrast_curve': lambda s: _get_curve(4.5) if s.platform == 'phone' else _get_curve(7),
        })
        return extend_spec_version(base, '2025', color2025)

    #################################################################
    # Remapped Colors                                             #
    #################################################################

    def surface_variant(self) -> 'DynamicColor':
        base = super().surface_variant()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_variant',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: self.surface_container_highest().get_tone(s),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def surface_tint(self) -> 'DynamicColor':
        base = super().surface_tint()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'surface_tint',
            'palette': lambda s: s.primary_palette,
            'tone': lambda s: self.primary().get_tone(s),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def background(self) -> 'DynamicColor':
        base = super().background()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'background',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: self.surface().get_tone(s),
            'is_background': True,
        })
        return extend_spec_version(base, '2025', color2025)

    def on_background(self) -> 'DynamicColor':
        base = super().on_background()
        color2025: DynamicColor = DynamicColor.from_palette({
            'name': 'on_background',
            'palette': lambda s: s.neutral_palette,
            'tone': lambda s: self.on_surface().get_tone(s),
            'background': lambda s: self.background(),
            'contrast_curve': lambda s: self.on_surface().contrast_curve(s),
        })
        return extend_spec_version(base, '2025', color2025)


# temperature/temperature_cache.py

"""
Design utilities using color temperature theory.

Analogous colors, complementary color, and cache to efficiently, lazily,
generate data for calculations when needed.
"""

import math
from typing import List, Optional, Mapping

from PyMCUlib.hct.hct import Hct
from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils


class TemperatureCache:
    """
    Design utilities using color temperature theory.

    Analogous colors, complementary color, and cache to efficiently, lazily,
    generate data for calculations when needed.
    """

    def __init__(self, input_color: Hct):
        """
        Initialize with an input color.

        Args:
            input_color: The input color.
        """
        self.input = input_color
        self.hcts_by_temp_cache: List[Hct] = []
        self.hcts_by_hue_cache: List[Hct] = []
        self.temps_by_hct_cache: Mapping[Hct, float] = {}
        self.input_relative_temperature_cache: float = -1.0
        self.complement_cache: Optional[Hct] = None

    @property
    def hcts_by_temp(self) -> List[Hct]:
        """
        HCTs for all hues, with the same chroma/tone as the input,
        sorted by temperature.

        Returns:
            List of Hct colors, sorted by temperature.
        """
        if len(self.hcts_by_temp_cache) > 0:
            return self.hcts_by_temp_cache

        hcts = self.hcts_by_hue + [self.input]
        temperatures_by_hct = self.temps_by_hct
        hcts.sort(key=lambda hct: temperatures_by_hct[hct])
        self.hcts_by_temp_cache = hcts
        return hcts

    @property
    def warmest(self) -> Hct:
        """
        Returns the warmest color with the same chroma/tone as the input.

        Returns:
            The warmest color.
        """
        return self.hcts_by_temp[-1]

    @property
    def coldest(self) -> Hct:
        """
        Returns the coldest color with the same chroma/tone as the input.

        Returns:
            The coldest color.
        """
        return self.hcts_by_temp[0]

    def analogous(self, count: int = 5, divisions: int = 12) -> List[Hct]:
        """
        A set of colors with differing hues, equidistant in temperature.

        In art, this is usually described as a set of 5 colors on a color wheel
        divided into 12 sections. This method allows provision of either of those
        values.

        Behavior is undefined when count or divisions is 0.
        When divisions < count, colors repeat.

        Args:
            count: The number of colors to return, includes the input color.
            divisions: The number of divisions on the color wheel.

        Returns:
            A list of count colors analogous to the input color.
        """
        start_hue = round(self.input.hue)
        start_hct = self.hcts_by_hue[start_hue]
        last_temp = self.relative_temperature(start_hct)
        all_colors = [start_hct]

        absolute_total_temp_delta = 0.0
        for i in range(360):
            hue = math_utils.sanitize_degrees_int(start_hue + i)
            hct = self.hcts_by_hue[hue]
            temp = self.relative_temperature(hct)
            temp_delta = abs(temp - last_temp)
            last_temp = temp
            absolute_total_temp_delta += temp_delta

        hue_addend = 1
        temp_step = absolute_total_temp_delta / divisions
        total_temp_delta = 0.0
        last_temp = self.relative_temperature(start_hct)

        while len(all_colors) < divisions:
            hue = math_utils.sanitize_degrees_int(start_hue + hue_addend)
            hct = self.hcts_by_hue[hue]
            temp = self.relative_temperature(hct)
            temp_delta = abs(temp - last_temp)
            total_temp_delta += temp_delta

            desired_total_temp_delta_for_index = len(all_colors) * temp_step
            index_satisfied = total_temp_delta >= desired_total_temp_delta_for_index
            index_addend = 1
            # Keep adding this hue to the answers until its temperature is
            # insufficient. This ensures consistent behavior when there aren't
            # [divisions] discrete steps between 0 and 360 in hue with [tempStep]
            # delta in temperature between them.
            #
            # For example, white and black have no analogues: there are no other
            # colors at T100/T0. Therefore, they should just be added to the array
            # as answers.
            while index_satisfied and len(all_colors) < divisions:
                all_colors.append(hct)
                desired_total_temp_delta_for_index = (
                    (len(all_colors) + index_addend) * temp_step
                )
                index_satisfied = total_temp_delta >= desired_total_temp_delta_for_index
                index_addend += 1

            last_temp = temp
            hue_addend += 1
            if hue_addend > 360:
                while len(all_colors) < divisions:
                    all_colors.append(hct)
                break

        answers = [self.input]

        # First, generate analogues from rotating counter-clockwise.
        increase_hue_count = math.floor((count - 1) / 2.0)
        for i in range(1, increase_hue_count + 1):
            index = 0 - i
            while index < 0:
                index = len(all_colors) + index
            if index >= len(all_colors):
                index = index % len(all_colors)
            answers.insert(0, all_colors[index])

        # Second, generate analogues from rotating clockwise.
        decrease_hue_count = count - increase_hue_count - 1
        for i in range(1, decrease_hue_count + 1):
            index = i
            while index < 0:
                index = len(all_colors) + index
            if index >= len(all_colors):
                index = index % len(all_colors)
            answers.append(all_colors[index])

        return answers

    @property
    def complement(self) -> Hct:
        """
        A color that complements the input color aesthetically.

        In art, this is usually described as being across the color wheel.
        History of this shows intent as a color that is just as cool-warm as the
        input color is warm-cool.

        Returns:
            The complementary color.
        """
        if self.complement_cache is not None:
            return self.complement_cache

        coldest_hue = self.coldest.hue
        coldest_temp = self.temps_by_hct[self.coldest]

        warmest_hue = self.warmest.hue
        warmest_temp = self.temps_by_hct[self.warmest]
        range_val = warmest_temp - coldest_temp
        # If the coldest and warmest temperatures are the same, range_val is 0,
        # return the input color to avoid division by 0.
        if range_val == 0.0:
            self.complement_cache = self.input
            return self.input

        start_hue_is_coldest_to_warmest = TemperatureCache._is_between(
            self.input.hue, coldest_hue, warmest_hue
        )
        start_hue = warmest_hue if start_hue_is_coldest_to_warmest else coldest_hue
        end_hue = coldest_hue if start_hue_is_coldest_to_warmest else warmest_hue
        direction_of_rotation = 1.0
        smallest_error = 1000.0
        answer = self.hcts_by_hue[round(self.input.hue)]

        complement_relative_temp = 1.0 - self.input_relative_temperature
        # Find the color in the other section, closest to the inverse percentile
        # of the input color. This is the complement.
        for hue_addend in range(361):
            hue = math_utils.sanitize_degrees_double(
                start_hue + direction_of_rotation * hue_addend
            )
            if not TemperatureCache._is_between(hue, start_hue, end_hue):
                continue
            possible_answer = self.hcts_by_hue[round(hue)]
            relative_temp = (
                self.temps_by_hct[possible_answer] - coldest_temp
            ) / range_val
            error = abs(complement_relative_temp - relative_temp)
            if error < smallest_error:
                smallest_error = error
                answer = possible_answer

        self.complement_cache = answer
        return self.complement_cache

    def relative_temperature(self, hct: Hct) -> float:
        """
        Temperature relative to all colors with the same chroma and tone.
        Value on a scale from 0 to 1.

        Args:
            hct: The color to get the relative temperature of.

        Returns:
            The relative temperature of the color, from 0 to 1.
        """
        range_val = (
            self.temps_by_hct[self.warmest] - self.temps_by_hct[self.coldest]
        )
        difference_from_coldest = (
            self.temps_by_hct[hct] - self.temps_by_hct[self.coldest]
        )
        # Handle when there's no difference in temperature between warmest and
        # coldest: for example, at T100, only one color is available, white.
        if range_val == 0.0:
            return 0.5
        return difference_from_coldest / range_val

    @property
    def input_relative_temperature(self) -> float:
        """
        Relative temperature of the input color. See relativeTemperature.

        Returns:
            The relative temperature of the input color.
        """
        if self.input_relative_temperature_cache >= 0.0:
            return self.input_relative_temperature_cache

        self.input_relative_temperature_cache = self.relative_temperature(self.input)
        return self.input_relative_temperature_cache

    @property
    def temps_by_hct(self) -> Mapping[Hct, float]:
        """
        A Dict with keys of HCTs in hctsByTemp, values of raw temperature.

        Returns:
            A Dict mapping HCT colors to their raw temperature.
        """
        if len(self.temps_by_hct_cache) > 0:
            return self.temps_by_hct_cache

        all_hcts = self.hcts_by_hue + [self.input]
        temperatures_by_hct = {}
        for hct in all_hcts:
            temperatures_by_hct[hct] = TemperatureCache.raw_temperature(hct)
        self.temps_by_hct_cache = temperatures_by_hct
        return self.temps_by_hct_cache

    @property
    def hcts_by_hue(self) -> List[Hct]:
        """
        HCTs for all hues, with the same chroma/tone as the input.
        Sorted ascending, hue 0 to 360.

        Returns:
            List of Hct colors, sorted by hue.
        """
        if len(self.hcts_by_hue_cache) > 0:
            return self.hcts_by_hue_cache

        hcts = []
        for hue in range(361):
            color_at_hue = Hct.from_hct(hue, self.input.chroma, self.input.tone)
            hcts.append(color_at_hue)
        self.hcts_by_hue_cache = hcts
        return self.hcts_by_hue_cache

    @staticmethod
    def _is_between(angle: float, a: float, b: float) -> bool:
        """
        Determines if an angle is between two other angles, rotating clockwise.

        Args:
            angle: The angle to check.
            a: The first angle.
            b: The second angle.

        Returns:
            True if the angle is between the two angles, false otherwise.
        """
        if a < b:
            return a <= angle and angle <= b
        return a <= angle or angle <= b

    @staticmethod
    def raw_temperature(color: Hct) -> float:
        """
        Value representing cool-warm factor of a color.
        Values below 0 are considered cool, above, warm.

        Color science has researched emotion and harmony, which art uses to select
        colors. Warm-cool is the foundation of analogous and complementary colors.
        See:
        - Li-Chen Ou's Chapter 19 in Handbook of Color Psychology (2015).
        - Josef Albers' Interaction of Color chapters 19 and 21.

        Implementation of Ou, Woodcock and Wright's algorithm, which uses
        L*a*b* / LCH color space.
        Return value has these properties:
        - Values below 0 are cool, above 0 are warm.
        - Lower bound: -0.52 - (chroma ^ 1.07 / 20). L*a*b* chroma is infinite.
          Assuming max of 130 chroma, -9.66.
        - Upper bound: -0.52 + (chroma ^ 1.07 / 20). L*a*b* chroma is infinite.
          Assuming max of 130 chroma, 8.61.

        Args:
            color: The color to get the raw temperature of.

        Returns:
            The raw temperature of the color.
        """
        lab = color_utils.lab_from_argb(color.to_int())
        hue = math_utils.sanitize_degrees_double(
            math.atan2(lab[2], lab[1]) * 180.0 / math.pi
        )
        chroma = math.sqrt((lab[1] * lab[1]) + (lab[2] * lab[2]))
        temperature = (
            -0.5
            + 0.02
            * math.pow(chroma, 1.07)
            * math.cos(math_utils.sanitize_degrees_double(hue - 50.0) * math.pi / 180.0)
        )
        return temperature


# utils/image_utils.py

from typing import List, Union, Optional, Tuple
import numpy as np
from PIL import Image
import io
from PyMCUlib.utils import color_utils


def source_color_from_image(image_data: Union[str, bytes, Image.Image],
                            area: Optional[Tuple[int, int, int, int]] = None) -> int:
    """
    Get the source color from an image.

    Args:
        image_data: The image data, can be a file path, bytes, or PIL Image object
        area: Optional crop area as (left, top, width, height)

    Returns:
        Source color - the color most suitable for creating a UI theme
    """
    # Convert to PIL Image if not already
    if isinstance(image_data, str):
        # Assume it's a file path
        img = Image.open(image_data)
    elif isinstance(image_data, bytes):
        # Bytes data
        img = Image.open(io.BytesIO(image_data))
    elif isinstance(image_data, Image.Image):
        # Already a PIL Image
        img = image_data
    else:
        raise ValueError("Unsupported image data type")
    
    # Determine crop rectangle
    if area is not None:
        sx, sy, sw, sh = area
        rect = (sx, sy, sx + sw, sy + sh)
    else:
        rect = (0, 0, img.width, img.height)
    
    # Convert to RGBA if not already
    if img.mode != 'RGBA':
        img = img.convert('RGBA')
    
    # Get image data
    cropped_img = img.crop(rect)
    img_array = np.array(cropped_img)
    
    # Convert to flattened bytes
    image_bytes = img_array.flatten()
    
    return source_color_from_image_bytes(image_bytes)


def source_color_from_image_bytes(image_bytes: np.ndarray) -> int:
    """
    Get the source color from image bytes.

    Args:
        image_bytes: The image bytes as a numpy array

    Returns:
        Source color - the color most suitable for creating a UI theme
    """
    # Convert Image data to Pixel Array
    pixels: List[int] = []
    
    # Handle the data in RGBA format (R,G,B,A sequence)
    for i in range(0, len(image_bytes), 4):
        if i + 3 >= len(image_bytes):
            break
            
        r = int(image_bytes[i])
        g = int(image_bytes[i + 1])
        b = int(image_bytes[i + 2])
        a = int(image_bytes[i + 3])
        
        if a < 255:
            continue
            
        argb = color_utils.argb_from_rgb(r, g, b)
        pixels.append(argb)
    
    # Convert Pixels to Material Colors
    from PyMCUlib.quantize.quantizer_celebi import QuantizerCelebi
    result = QuantizerCelebi.quantize(pixels, 128)
    from PyMCUlib.score.score import Score
    ranked = Score.score(result)
    top = ranked[0]
    return top

def source_color_from_file(file_path: str,
                            area: Optional[Tuple[int, int, int, int]] = None) -> int:
    """
    Get the source color from an image file.

    Args:
        file_path: Path to the image file
        area: Optional crop area as (left, top, width, height)

    Returns:
        Source color - the color most suitable for creating a UI theme

    Raises:
        FileNotFoundError: If the file does not exist
    """
    image = Image.open(file_path)
    return source_color_from_image(image, area)


# score/score.py

from typing import List, Mapping, Optional, TypedDict
from PyMCUlib.hct.hct import Hct
from PyMCUlib.utils import math_utils

class ScoreOptions(TypedDict, total=False):
    """
    Default options for ranking colors based on usage counts.
    desired: is the max count of the colors returned.
    fallbackColorARGB: Is the default color that should be used if no
                      other colors are suitable.
    filter: controls if the resulting colors should be filtered to not include
           hues that are not used often enough, and colors that are effectively
           grayscale.
    """
    desired: int
    fallbackColorARGB: int
    filter: bool

SCORE_OPTION_DEFAULTS = {
    "desired": 4,  # 4 colors matches what Android wallpaper picker.
    "fallbackColorARGB": 0xff4285f4,  # Google Blue.
    "filter": True,  # Avoid unsuitable colors.
}

class Score:
    """
    Given a large set of colors, remove colors that are unsuitable for a UI
    theme, and rank the rest based on suitability.
    
    Enables use of a high cluster count for image quantization, thus ensuring
    colors aren't muddied, while curating the high cluster count to a much
    smaller number of appropriate choices.
    """
    
    TARGET_CHROMA = 48.0  # A1 Chroma
    WEIGHT_PROPORTION = 0.7
    WEIGHT_CHROMA_ABOVE = 0.3
    WEIGHT_CHROMA_BELOW = 0.1
    CUTOFF_CHROMA = 5.0
    CUTOFF_EXCITED_PROPORTION = 0.01
    
    def __init__(self):
        """Score cannot be instantiated."""
        raise NotImplementedError("Score cannot be instantiated")
    
    @classmethod
    def score(
        cls,
        colors_to_population: Mapping[int, int],
        options: Optional[ScoreOptions] = None
    ) -> List[int]:
        """
        Given a map with keys of colors and values of how often the color appears,
        rank the colors based on suitability for being used for a UI theme.
        
        Args:
            colors_to_population: map with keys of colors and values of how often
                the color appears, usually from a source image.
            options: optional parameters.
                
        Returns:
            Colors sorted by suitability for a UI theme. The most suitable
            color is the first item, the least suitable is the last. There will
            always be at least one color returned. If all the input colors
            were not suitable for a theme, a default fallback color will be
            provided, Google Blue.
        """
        # Apply default options if not provided
        if options is None:
            options = {}
        opts = {**SCORE_OPTION_DEFAULTS, **options}
        desired = opts["desired"]
        fallback_color_argb = opts["fallbackColorARGB"]
        filter_colors = opts["filter"]
        
        # Get the HCT color for each Argb value, while finding the per hue count and
        # total count.
        colors_hct = []
        hue_population = [0] * 360
        population_sum = 0
        
        for argb, population in colors_to_population.items():
            hct = Hct.from_int(argb)
            colors_hct.append(hct)
            hue = int(hct.hue)
            hue_population[hue] += population
            population_sum += population
        
        # Hues with more usage in neighboring 30 degree slice get a larger number.
        hue_excited_proportions = [0.0] * 360
        for hue in range(360):
            proportion = hue_population[hue] / population_sum if population_sum > 0 else 0
            for i in range(hue - 14, hue + 16):
                neighbor_hue = math_utils.sanitize_degrees_int(i)
                hue_excited_proportions[neighbor_hue] += proportion
        
        # Scores each HCT color based on usage and chroma, while optionally
        # filtering out values that do not have enough chroma or usage.
        scored_hct = []
        for hct in colors_hct:
            hue = math_utils.sanitize_degrees_int(round(hct.hue))
            proportion = hue_excited_proportions[hue]
            if filter_colors and (hct.chroma < cls.CUTOFF_CHROMA or 
                                 proportion <= cls.CUTOFF_EXCITED_PROPORTION):
                continue
            
            proportion_score = proportion * 100.0 * cls.WEIGHT_PROPORTION
            chroma_weight = cls.WEIGHT_CHROMA_BELOW if hct.chroma < cls.TARGET_CHROMA else cls.WEIGHT_CHROMA_ABOVE
            chroma_score = (hct.chroma - cls.TARGET_CHROMA) * chroma_weight
            score = proportion_score + chroma_score
            scored_hct.append({"hct": hct, "score": score})
        
        # Sort so that colors with higher scores come first
        scored_hct.sort(key=lambda x: x["score"], reverse=True)
        
        # Iterates through potential hue differences in degrees in order to select
        # the colors with the largest distribution of hues possible. Starting at
        # 90 degrees(maximum difference for 4 colors) then decreasing down to a
        # 15 degree minimum.
        chosen_colors = []
        for difference_degrees in range(90, 14, -1):
            chosen_colors.clear()
            for entry in scored_hct:
                hct = entry["hct"]
                duplicate_hue = False
                for chosen_hct in chosen_colors:
                    if math_utils.difference_degrees(hct.hue, chosen_hct.hue) < difference_degrees:
                        duplicate_hue = True
                        break
                if not duplicate_hue:
                    chosen_colors.append(hct)
                if len(chosen_colors) >= desired:
                    break
            if len(chosen_colors) >= desired:
                break
        
        colors = []
        if len(chosen_colors) == 0:
            colors.append(fallback_color_argb)
        
        for chosen_hct in chosen_colors:
            colors.append(chosen_hct.to_int())
        
        return colors


# quantize/point_provider.py

from abc import ABC, abstractmethod
from typing import List


class PointProvider(ABC):
    """
    An interface to allow use of different color spaces by
    quantizers.
    """

    @abstractmethod
    def to_int(self, point: List[float]) -> int:
        """
        Converts a point in a color space to an ARGB integer.
        
        Args:
            point: Color coordinates in the color space
            
        Returns:
            The color represented as an ARGB integer
        """
        pass

    @abstractmethod
    def from_int(self, argb: int) -> List[float]:
        """
        Converts an ARGB integer to a point in a color space.
        
        Args:
            argb: The color represented as an ARGB integer
            
        Returns:
            Color coordinates in the color space
        """
        pass

    @abstractmethod
    def distance(self, from_point: List[float], to_point: List[float]) -> float:
        """
        Calculates the distance between two points in a color space.
        
        Args:
            from_point: The first point in the color space
            to_point: The second point in the color space
            
        Returns:
            The distance between the two points
        """
        pass


# quantize/lab_point_provider.py

from typing import List
from PyMCUlib.utils import color_utils
from PyMCUlib.quantize.point_provider import PointProvider


class LabPointProvider(PointProvider):
    """
    Provides conversions needed for K-Means quantization. Converting input to
    points, and converting the final state of the K-Means algorithm to colors.
    """

    def from_int(self, argb: int) -> List[float]:
        """
        Convert a color represented in ARGB to a 3-element array of L*a*b*
        coordinates of the color.
        
        Args:
            argb: The ARGB color integer
            
        Returns:
            A list of L*, a*, b* coordinates
        """
        return color_utils.lab_from_argb(argb)

    def to_int(self, point: List[float]) -> int:
        """
        Convert a 3-element array to a color represented in ARGB.
        
        Args:
            point: A list containing L*, a*, b* coordinates
            
        Returns:
            The ARGB color integer
        """
        return color_utils.argb_from_lab(point[0], point[1], point[2])

    def distance(self, from_point: List[float], to_point: List[float]) -> float:
        """
        Standard CIE 1976 delta E formula also takes the square root, unneeded
        here. This method is used by quantization algorithms to compare distance,
        and the relative ordering is the same, with or without a square root.
        
        This relatively minor optimization is helpful because this method is
        called at least once for each pixel in an image.
        
        Args:
            from_point: The first point in Lab color space
            to_point: The second point in Lab color space
            
        Returns:
            The squared distance between the two points
        """
        d_l = from_point[0] - to_point[0]
        d_a = from_point[1] - to_point[1]
        d_b = from_point[2] - to_point[2]
        return d_l * d_l + d_a * d_a + d_b * d_b


# scheme/__init__.py

# --- scheme ---
from PyMCUlib.scheme.scheme import Scheme

# --- scheme_android ---
from PyMCUlib.scheme.scheme_android import SchemeAndroid

# --- scheme_neutral ---
from PyMCUlib.scheme.scheme_neutral import SchemeNeutral

# --- scheme_expressive ---
from PyMCUlib.scheme.scheme_expressive import SchemeExpressive

# --- scheme_tonal_spot ---
from PyMCUlib.scheme.scheme_tonal_spot import SchemeTonalSpot

# --- scheme_rainbow ---
from PyMCUlib.scheme.scheme_rainbow import SchemeRainbow

# --- scheme_content ---
from PyMCUlib.scheme.scheme_content import SchemeContent

# --- scheme_fidelity ---
from PyMCUlib.scheme.scheme_fidelity import SchemeFidelity

# --- scheme_vibrant ---
from PyMCUlib.scheme.scheme_vibrant import SchemeVibrant

# --- scheme_monochrome ---
from PyMCUlib.scheme.scheme_monochrome import SchemeMonochrome

# --- scheme_fruit_salad ---
from PyMCUlib.scheme.scheme_fruit_salad import SchemeFruitSalad

__all__ = [
    # scheme
    "Scheme",
    # scheme_android
    "SchemeAndroid",
    # scheme_neutral
    "SchemeNeutral",
    # scheme_expressive
    "SchemeExpressive",
    # scheme_tonal_spot
    "SchemeTonalSpot",
    # scheme_rainbow
    "SchemeRainbow",
    # scheme_content
    "SchemeContent",
    # scheme_fidelity
    "SchemeFidelity",
    # scheme_vibrant
    "SchemeVibrant",
    # scheme_monochrome
    "SchemeMonochrome",
    # scheme_fruit_salad
    "SchemeFruitSalad",
]


# quantize/quantizer_celebi.py

from typing import Dict, List
from PyMCUlib.quantize.quantizer_wsmeans import QuantizerWsmeans
from PyMCUlib.quantize.quantizer_wu import QuantizerWu


class QuantizerCelebi:
    """
    An image quantizer that improves on the quality of a standard K-Means
    algorithm by setting the K-Means initial state to the output of a Wu
    quantizer, instead of random centroids. Improves on speed by several
    optimizations, as implemented in Wsmeans, or Weighted Square Means, K-Means
    with those optimizations.

    This algorithm was designed by M. Emre Celebi, and was found in their 2011
    paper, Improving the Performance of K-Means for Color Quantization.
    https://arxiv.org/abs/1101.0395
    """

    @staticmethod
    def quantize(pixels: List[int], max_colors: int) -> Dict[int, int]:
        """
        Args:
            pixels: Colors in ARGB format.
            max_colors: The number of colors to divide the image into. A lower
                number of colors may be returned.
        
        Returns:
            Dict with keys of colors in ARGB format, and values of number of
            pixels in the original image that correspond to the color in the
            quantized image.
        """
        wu = QuantizerWu()
        wu_result = wu.quantize(pixels, max_colors)
        return QuantizerWsmeans.quantize(pixels, wu_result, max_colors)


# dynamiccolor/contrast_curve.py

"""
A class containing a value that changes with the contrast level.

Usually represents the contrast requirements for a dynamic color on its
background. The four values correspond to values for contrast levels -1.0,
0.0, 0.5, and 1.0, respectively.
"""

from PyMCUlib.utils import math_utils

class ContrastCurve:
    """
    A class containing a value that changes with the contrast level.

    Usually represents the contrast requirements for a dynamic color on its
    background. The four values correspond to values for contrast levels -1.0,
    0.0, 0.5, and 1.0, respectively.
    """

    def __init__(self, low: float, normal: float, medium: float, high: float):
        """
        Creates a `ContrastCurve` object.

        Args:
            low: Value for contrast level -1.0
            normal: Value for contrast level 0.0
            medium: Value for contrast level 0.5
            high: Value for contrast level 1.0
        """
        self.low = low
        self.normal = normal
        self.medium = medium
        self.high = high

    def get(self, contrast_level: float) -> float:
        """
        Returns the value at a given contrast level.

        Args:
            contrast_level: The contrast level. 0.0 is the default (normal); -1.0
                is the lowest; 1.0 is the highest.

        Returns:
            The value. For contrast ratios, a number between 1.0 and 21.0.
        """
        if contrast_level <= -1.0:
            return self.low
        elif contrast_level < 0.0:
            return math_utils.lerp(self.low, self.normal, (contrast_level - (-1)) / 1)
        elif contrast_level < 0.5:
            return math_utils.lerp(self.normal, self.medium, (contrast_level - 0) / 0.5)
        elif contrast_level < 1.0:
            return math_utils.lerp(self.medium, self.high, (contrast_level - 0.5) / 0.5)
        else:
            return self.high


# dynamiccolor/color_spec_delegate.py

from abc import ABC, abstractmethod
from typing import Literal, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor
    from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme

# define specification version type
SpecVersion = Literal['2021', '2025']


class ColorSpecDelegate(ABC):
    """
    A delegate that provides the dynamic color constraints for
    MaterialDynamicColors.

    This is used to allow for different color constraints for different spec
    versions.
    """

    ################################################################
    # Main Palettes                                              #
    ################################################################

    @abstractmethod
    def primary_palette_key_color(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def secondary_palette_key_color(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def tertiary_palette_key_color(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def neutral_palette_key_color(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def neutral_variant_palette_key_color(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def error_palette_key_color(self) -> 'DynamicColor':
        pass

    ################################################################
    # Surfaces [S]                                               #
    ################################################################

    @abstractmethod
    def background(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_background(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_dim(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_bright(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_container_lowest(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_container_low(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_container_high(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_container_highest(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_surface(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_variant(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_surface_variant(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def inverse_surface(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def inverse_on_surface(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def outline(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def outline_variant(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def shadow(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def scrim(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def surface_tint(self) -> 'DynamicColor':
        pass

    ################################################################
    # Primaries [P]                                              #
    ################################################################

    @abstractmethod
    def primary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def primary_dim(self) -> Optional['DynamicColor']:
        pass

    @abstractmethod
    def on_primary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def primary_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_primary_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def inverse_primary(self) -> 'DynamicColor':
        pass

    ################################################################
    # Secondaries [Q]                                            #
    ################################################################

    @abstractmethod
    def secondary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def secondary_dim(self) -> Optional['DynamicColor']:
        pass

    @abstractmethod
    def on_secondary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def secondary_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_secondary_container(self) -> 'DynamicColor':
        pass

    ################################################################
    # Tertiaries [T]                                             #
    ################################################################

    @abstractmethod
    def tertiary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def tertiary_dim(self) -> Optional['DynamicColor']:
        pass

    @abstractmethod
    def on_tertiary(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def tertiary_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_tertiary_container(self) -> 'DynamicColor':
        pass

    ################################################################
    # Errors [E]                                                 #
    ################################################################

    @abstractmethod
    def error(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def error_dim(self) -> Optional['DynamicColor']:
        pass

    @abstractmethod
    def on_error(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def error_container(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_error_container(self) -> 'DynamicColor':
        pass

    ################################################################
    # Primary Fixed Colors [PF]                                  #
    ################################################################

    @abstractmethod
    def primary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def primary_fixed_dim(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_primary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_primary_fixed_variant(self) -> 'DynamicColor':
        pass

    ################################################################
    # Secondary Fixed Colors [QF]                                #
    ################################################################

    @abstractmethod
    def secondary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def secondary_fixed_dim(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_secondary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_secondary_fixed_variant(self) -> 'DynamicColor':
        pass

    ################################################################
    # Tertiary Fixed Colors [TF]                                 #
    ################################################################

    @abstractmethod
    def tertiary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def tertiary_fixed_dim(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_tertiary_fixed(self) -> 'DynamicColor':
        pass

    @abstractmethod
    def on_tertiary_fixed_variant(self) -> 'DynamicColor':
        pass

    ################################################################
    # Other                                                      #
    ################################################################

    @abstractmethod
    def highest_surface(self, s: 'DynamicScheme') -> 'DynamicColor':
        pass


# palettes/tonal_palette.py

"""
A convenience class for retrieving colors that are constant in hue and
chroma, but vary in tone.
"""

from PyMCUlib.hct.hct import Hct


class TonalPalette:
    """
    A convenience class for retrieving colors that are constant in hue and
    chroma, but vary in tone.
    """

    def __init__(self, hue: float, chroma: float, key_color: Hct) -> None:
        """
        Initialize a TonalPalette with specific hue, chroma, and key color.
        
        Args:
            hue: HCT hue
            chroma: HCT chroma
            key_color: Hct object that represents the key color
        """
        self.hue = hue
        self.chroma = chroma
        self.key_color = key_color
        self._cache = {}

    @staticmethod
    def from_int(argb: int) -> 'TonalPalette':
        """
        Create a TonalPalette from an ARGB integer.
        
        Args:
            argb: ARGB representation of a color
            
        Returns:
            Tones matching that color's hue and chroma.
        """
        hct = Hct.from_int(argb)
        return TonalPalette.from_hct(hct)

    @staticmethod
    def from_hct(hct: Hct) -> 'TonalPalette':
        """
        Create a TonalPalette from an Hct object.
        
        Args:
            hct: Hct object
            
        Returns:
            Tones matching that color's hue and chroma.
        """
        return TonalPalette(hct.hue, hct.chroma, hct)

    @staticmethod
    def from_hue_and_chroma(hue: float, chroma: float) -> 'TonalPalette':
        """
        Create a TonalPalette from hue and chroma values.
        
        Args:
            hue: HCT hue
            chroma: HCT chroma
            
        Returns:
            Tones matching hue and chroma.
        """
        key_color = KeyColor(hue, chroma).create()
        return TonalPalette(hue, chroma, key_color)

    def tone(self, tone: float) -> int:
        """
        Get the ARGB representation of a color with specified tone.
        
        Args:
            tone: HCT tone, measured from 0 to 100.
            
        Returns:
            ARGB representation of a color with that tone.
        """
        if tone not in self._cache:
            self._cache[tone] = Hct.from_hct(self.hue, self.chroma, tone).to_int()
        return self._cache[tone]

    def get_hct(self, tone: float) -> Hct:
        """
        Get the HCT representation of a color with specified tone.
        
        Args:
            tone: HCT tone.
            
        Returns:
            HCT representation of a color with that tone.
        """
        return Hct.from_int(self.tone(tone))


class KeyColor:
    """
    Key color is a color that represents the hue and chroma of a tonal palette
    """

    def __init__(self, hue: float, requested_chroma: float) -> None:
        """
        Initialize a KeyColor with specific hue and requested chroma.
        
        Args:
            hue: HCT hue
            requested_chroma: HCT chroma
        """
        self.hue = hue
        self.requested_chroma = requested_chroma
        self._chroma_cache = {}
        self._max_chroma_value = 200.0

    def create(self) -> Hct:
        """
        Creates a key color from a hue and a chroma.
        The key color is the first tone, starting from T50, matching the given hue
        and chroma.
        
        Returns:
            Key color Hct
        """
        # Pivot around T50 because T50 has the most chroma available, on
        # average. Thus it is most likely to have a direct answer.
        pivot_tone = 50
        tone_step_size = 1
        # Epsilon to accept values slightly higher than the requested chroma.
        epsilon = 0.01

        # Binary search to find the tone that can provide a chroma that is closest
        # to the requested chroma.
        lower_tone = 0
        upper_tone = 100
        while lower_tone < upper_tone:
            mid_tone = int((lower_tone + upper_tone) / 2)
            is_ascending = self._max_chroma(mid_tone) < self._max_chroma(mid_tone + tone_step_size)
            sufficient_chroma = self._max_chroma(mid_tone) >= self.requested_chroma - epsilon

            if sufficient_chroma:
                # Either range [lower_tone, mid_tone] or [mid_tone, upper_tone] has
                # the answer, so search in the range that is closer the pivot tone.
                if abs(lower_tone - pivot_tone) < abs(upper_tone - pivot_tone):
                    upper_tone = mid_tone
                else:
                    if lower_tone == mid_tone:
                        return Hct.from_hct(self.hue, self.requested_chroma, lower_tone)
                    lower_tone = mid_tone
            else:
                # As there is no sufficient chroma in the mid_tone, follow the direction
                # to the chroma peak.
                if is_ascending:
                    lower_tone = mid_tone + tone_step_size
                else:
                    # Keep mid_tone for potential chroma peak.
                    upper_tone = mid_tone

        return Hct.from_hct(self.hue, self.requested_chroma, lower_tone)

    def _max_chroma(self, tone: float) -> float:
        """
        Find the maximum chroma for a given tone.
        
        Args:
            tone: HCT tone
            
        Returns:
            Maximum available chroma for the given tone
        """
        if tone in self._chroma_cache:
            return self._chroma_cache[tone]
        
        chroma = Hct.from_hct(self.hue, self._max_chroma_value, tone).chroma
        self._chroma_cache[tone] = chroma
        return chroma


"""
Material Color Utilities Python Library

A Python library for color operations and conversions, providing the HCT color space,
CAM16 color appearance model, and color blending tools.
"""

__version__ = "1.0.2"

# --- blend ---
from PyMCUlib.blend import Blend

# --- contrast ---
from PyMCUlib.contrast import Contrast

# --- dislike ---
from PyMCUlib.dislike import DislikeAnalyzer

# --- dynamiccolor ---
from PyMCUlib.dynamiccolor import (
    ColorSpecDelegate,
    SpecVersion,
    get_spec,
    ContrastCurve,
    DynamicColor,
    extend_spec_version,
    validate_extended_color,
    DynamicScheme,
    MaterialDynamicColors,
    ToneDeltaPair,
    DeltaConstraint,
    TonePolarity,
    Variant,
)

# --- hct ---
from PyMCUlib.hct import ViewingConditions, Cam16, HctSolver, Hct

# --- palettes ---
from PyMCUlib.palettes import TonalPalette, KeyColor, CorePalette, CorePalettes

# --- quantize ---
from PyMCUlib.quantize import (
    PointProvider,
    LabPointProvider,
    QuantizerMap,
    QuantizerWu,
    QuantizerWsmeans,
    QuantizerCelebi,
)

# --- scheme ---
from PyMCUlib.scheme import (
    Scheme,
    SchemeAndroid,
    SchemeNeutral,
    SchemeExpressive,
    SchemeTonalSpot,
    SchemeRainbow,
    SchemeContent,
    SchemeFidelity,
    SchemeVibrant,
    SchemeMonochrome,
    SchemeFruitSalad,
)

# --- score ---
from PyMCUlib.score import Score, ScoreOptions

# --- temperature ---
from PyMCUlib.temperature import TemperatureCache

# --- utils ---
from PyMCUlib.utils import (
    argb_from_rgb,
    argb_from_linrgb,
    alpha_from_argb,
    red_from_argb,
    green_from_argb,
    blue_from_argb,
    is_opaque,
    argb_from_xyz,
    xyz_from_argb,
    argb_from_lab,
    lab_from_argb,
    argb_from_lstar,
    lstar_from_argb,
    y_from_lstar,
    lstar_from_y,
    linearized,
    delinearized,
    white_point_d65,
    hex_from_argb,
    argb_from_hex,
    source_color_from_image,
    source_color_from_image_bytes,
    source_color_from_file,
    signum,
    lerp,
    clamp_int,
    clamp_double,
    sanitize_degrees_int,
    sanitize_degrees_double,
    rotation_direction,
    difference_degrees,
    matrix_multiply,
    CustomColor,
    ColorGroup,
    CustomColorGroup,
    Theme,
    ApplyThemeOptions,
    theme_from_source_color,
    theme_from_image,
    custom_color,
    apply_theme,
    set_scheme_properties,
)

__all__ = [
    # Version
    "__version__",

    # blend
    "Blend",

    # contrast
    "Contrast",

    # dislike
    "DislikeAnalyzer",

    # dynamiccolor
    "ColorSpecDelegate",
    "SpecVersion",
    "get_spec",
    "ContrastCurve",
    "DynamicColor",
    "extend_spec_version",
    "validate_extended_color",
    "DynamicScheme",
    "MaterialDynamicColors",
    "ToneDeltaPair",
    "DeltaConstraint",
    "TonePolarity",
    "Variant",

    # hct
    "ViewingConditions",
    "Cam16",
    "HctSolver",
    "Hct",

    # palettes
    "TonalPalette",
    "KeyColor",
    "CorePalette",
    "CorePalettes",

    # quantize
    "PointProvider",
    "LabPointProvider",
    "QuantizerMap",
    "QuantizerWu",
    "QuantizerWsmeans",
    "QuantizerCelebi",

    # scheme
    "Scheme",
    "SchemeAndroid",
    "SchemeNeutral",
    "SchemeExpressive",
    "SchemeTonalSpot",
    "SchemeRainbow",
    "SchemeContent",
    "SchemeFidelity",
    "SchemeVibrant",
    "SchemeMonochrome",
    "SchemeFruitSalad",

    # score
    "Score",
    "ScoreOptions",

    # temperature
    "TemperatureCache",

    # utils
    "argb_from_rgb",
    "argb_from_linrgb",
    "alpha_from_argb",
    "red_from_argb",
    "green_from_argb",
    "blue_from_argb",
    "is_opaque",
    "argb_from_xyz",
    "xyz_from_argb",
    "argb_from_lab",
    "lab_from_argb",
    "argb_from_lstar",
    "lstar_from_argb",
    "y_from_lstar",
    "lstar_from_y",
    "linearized",
    "delinearized",
    "white_point_d65",
    "hex_from_argb",
    "argb_from_hex",
    "source_color_from_image",
    "source_color_from_image_bytes",
    "source_color_from_file",
    "signum",
    "lerp",
    "clamp_int",
    "clamp_double",
    "sanitize_degrees_int",
    "sanitize_degrees_double",
    "rotation_direction",
    "difference_degrees",
    "matrix_multiply",
    "CustomColor",
    "ColorGroup",
    "CustomColorGroup",
    "Theme",
    "ApplyThemeOptions",
    "theme_from_source_color",
    "theme_from_image",
    "custom_color",
    "apply_theme",
    "set_scheme_properties",
]


# scheme/scheme_rainbow.py

"""
A playful theme - the source color's hue does not appear in the theme.
"""

from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme
from PyMCUlib.dynamiccolor.variant import Variant
from PyMCUlib.hct.hct import Hct


class SchemeRainbow(DynamicScheme):
    """A playful theme - the source color's hue does not appear in the theme."""
    
    DEFAULT_SPEC_VERSION = '2021'
    DEFAULT_PLATFORM = 'phone'

    def __init__(self, source_color_hct: Hct, is_dark: bool, contrast_level: float):
        """
        Initialize a rainbow theme with a source color.
        
        Args:
            source_color_hct: Source color in HCT color space.
            is_dark: Whether the theme is in dark mode.
            contrast_level: The contrast level of the theme.
        """
        super().__init__({
            'source_color_hct': source_color_hct,
            'variant': Variant.RAINBOW,
            'contrast_level': contrast_level,
            'is_dark': is_dark,
            'platform': SchemeRainbow.DEFAULT_PLATFORM,
            'spec_version': SchemeRainbow.DEFAULT_SPEC_VERSION,
        })


# blend/blend.py

"""
Functions for blending in HCT and CAM16.
"""

from PyMCUlib.hct.cam16 import Cam16
from PyMCUlib.hct.hct import Hct
from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils


class Blend:
    """
    Functions for blending in HCT and CAM16.
    """

    @staticmethod
    def harmonize(design_color: int, source_color: int) -> int:
        """
        Blend the design color's HCT hue towards the key color's HCT
        hue, in a way that leaves the original color recognizable and
        recognizably shifted towards the key color.

        Args:
            design_color: ARGB representation of an arbitrary color.
            source_color: ARGB representation of the main theme color.
        
        Returns:
            The design color with a hue shifted towards the
            system's color, a slightly warmer/cooler variant of the design
            color's hue.
        """
        from_hct = Hct.from_int(design_color)
        to_hct = Hct.from_int(source_color)
        difference_degrees = math_utils.difference_degrees(from_hct.hue, to_hct.hue)
        rotation_degrees = min(difference_degrees * 0.5, 15.0)
        output_hue = math_utils.sanitize_degrees_double(
            from_hct.hue +
            rotation_degrees * math_utils.rotation_direction(from_hct.hue, to_hct.hue))
        return Hct.from_hct(output_hue, from_hct.chroma, from_hct.tone).to_int()

    @staticmethod
    def hct_hue(from_color: int, to_color: int, amount: float) -> int:
        """
        Blends hue from one color into another. The chroma and tone of
        the original color are maintained.

        Args:
            from_color: ARGB representation of color
            to_color: ARGB representation of color
            amount: how much blending to perform; 0.0 >= and <= 1.0
        
        Returns:
            from_color, with a hue blended towards to_color. Chroma and tone
            are constant.
        """
        ucs = Blend.cam16_ucs(from_color, to_color, amount)
        ucs_cam = Cam16.from_int(ucs)
        from_cam = Cam16.from_int(from_color)
        blended = Hct.from_hct(
            ucs_cam.hue,
            from_cam.chroma,
            color_utils.lstar_from_argb(from_color),
        )
        return blended.to_int()

    @staticmethod
    def cam16_ucs(from_color: int, to_color: int, amount: float) -> int:
        """
        Blend in CAM16-UCS space.

        Args:
            from_color: ARGB representation of color
            to_color: ARGB representation of color
            amount: how much blending to perform; 0.0 >= and <= 1.0
        
        Returns:
            from_color, blended towards to_color. Hue, chroma, and tone will
            change.
        """
        from_cam = Cam16.from_int(from_color)
        to_cam = Cam16.from_int(to_color)
        from_j = from_cam.jstar
        from_a = from_cam.astar
        from_b = from_cam.bstar
        to_j = to_cam.jstar
        to_a = to_cam.astar
        to_b = to_cam.bstar
        jstar = from_j + (to_j - from_j) * amount
        astar = from_a + (to_a - from_a) * amount
        bstar = from_b + (to_b - from_b) * amount
        return Cam16.from_ucs(jstar, astar, bstar).to_int()